"""
Compute intensity histograms of the underlying images in this dataset.

TODO:

    - [ ] Migrate to kwcoco proper

    - [ ] Fix accumulation of floating point values

    - [ ] Handle nodata

    - [x] Rename to coco_spectra

CommandLine:
    smartwatch spectra --src special:watch-msi --show=True --stat=density
    smartwatch spectra --src special:photos --show=True --fill=False
    smartwatch spectra --src special:shapes8 --show=True --stat=count --cumulative=True --multiple=stack

    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    smartwatch spectra --src $DVC_DATA_DPATH/Drop6/data.kwcoco.zip --channels='red|green|blue|nir' --workers=11

    DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc
    KWCOCO_FPATH=$DVC_DPATH/Drop2-Aligned-TA1-2022-01/combo_L_nowv_vali.kwcoco.json
    smartwatch spectra --src $KWCOCO_FPATH --show=True --show=True --include_channels="forest|water|bare_ground"
"""
import ubelt as ub
import scriptconfig as scfg
import math

try:
    from xdev import profile
except ImportError:
    profile = ub.identity


class CocoSpectraConfig(scfg.DataConfig):
    """
    Updates image transforms in a kwcoco json file to align all videos to a
    target GSD.
    """
    __default__ = {
        'src': scfg.Value('data.kwcoco.json', help='input coco dataset', position=1),

        'dst': scfg.Value(None, help='if specified dump the figure to disk at this file path (e.g. with a jpg or png suffix)'),

        'show': scfg.Value(False, isflag=True, help='if True, do a plt.show()'),
        'draw': scfg.Value(True, help='if False disables all visualization and just print tables'),

        'workers': scfg.Value(0, help='number of io workers'),
        'mode': scfg.Value('process', help='type of parallelism'),

        'include_channels': scfg.Value(None, help='if specified can be | separated valid channels', alias=['channels']),
        'exclude_channels': scfg.Value(None, help='if specified can be | separated invalid channels'),

        'include_sensors': scfg.Value(None, help='if specified can be comma separated valid sensors'),
        'exclude_sensors': scfg.Value(None, help='if specified can be comma separated invalid sensors'),

        'max_images': scfg.Value(None, help='if given only sample this many images when computing statistics'),

        'valid_range': scfg.Value(None, help='Only include values within this range; specified as <min_val>:<max_val> e.g. (0:10000)'),

        'title': scfg.Value(None, help='Provide a title for the histogram figure'),

        # Histogram modifiers
        'kde': scfg.Value(True, help='if True compute a kernel density estimate to smooth the distribution'),
        'cumulative': scfg.Value(False, help='If True, plot the cumulative counts as bins increase.'),

        # 'bins': scfg.Value(256, help='Generic bin parameter that can be the name of a reference rule or the number of bins.'),
        'bins': scfg.Value('auto', help='Generic bin parameter that can be the name of a reference rule or the number of bins.'),

        'fill': scfg.Value(True, isflag=True, help='If True, fill in the space under the histogram.'),
        'element': scfg.Value('step', help='Visual representation of the histogram statistic.', choices=['bars', 'step', 'poly']),
        'multiple': scfg.Value('layer', choices=['layer', 'dodge', 'stack', 'fill']
                               , help='Approach to resolving multiple elements when semantic mapping creates subsets.'),

        'stat': scfg.Value('probability', choices={'count', 'frequency', 'density', 'probability'}, help=ub.paragraph(
            '''
            Aggregate statistic to compute in each bin.

            - ``count`` shows the number of observations
            - ``frequency`` shows the number of observations divided by the bin width
            - ``density`` normalizes counts so that the area of the histogram is 1
            - ``probability`` normalizes counts so that the sum of the bar heights is 1
            ''')),

        'select_images': scfg.Value(
            None, type=str, help=ub.paragraph(
                '''
                A json query (via the jq spec) that specifies which images
                belong in the subset. Note, this is a passed as the body of
                the following jq query format string to filter valid ids
                '.images[] | select({select_images}) | .id'.

                Examples for this argument are as follows:
                '.id < 3' will select all image ids less than 3.
                '.file_name | test(".*png")' will select only images with
                file names that end with png.
                '.file_name | test(".*png") | not' will select only images
                with file names that do not end with png.
                '.myattr == "foo"' will select only image dictionaries
                where the value of myattr is "foo".
                '.id < 3 and (.file_name | test(".*png"))' will select only
                images with id less than 3 that are also pngs.
                .myattr | in({"val1": 1, "val4": 1}) will take images
                where myattr is either val1 or val4.

                Requries the "jq" python library is installed.
                ''')),

        'select_videos': scfg.Value(
            None, help=ub.paragraph(
                '''
                A json query (via the jq spec) that specifies which videos
                belong in the subset. Note, this is a passed as the body of
                the following jq query format string to filter valid ids
                '.videos[] | select({select_images}) | .id'.

                Examples for this argument are as follows:
                '.name | startswith("foo")' will select only videos
                where the name starts with foo.

                Only applicable for dataset that contain videos.

                Requries the "jq" python library is installed.
                ''')),
    }


class HistAccum:
    """
    Helper to accumulate histograms
    """

    def __init__(self):
        self.accum = {}
        self.n = 0

    def update(self, data, sensor, channel):
        self.n += 1
        if sensor not in self.accum:
            self.accum[sensor] = {}
        # TODO: video / month
        sensor_accum = self.accum[sensor]
        if channel not in sensor_accum:
            sensor_accum[channel] = ub.ddict(lambda: 0)
        final_accum = sensor_accum[channel]
        for k, v in data.items():
            final_accum[k] += v

    def finalize(self):
        import pandas as pd
        import numpy as np
        # Stack all accuulated histograms into a longform dataframe
        to_stack = {}
        for sensor, sub in self.accum.items():
            for channel, hist in sub.items():
                hist = ub.sorted_keys(hist)
                # hist.pop(0)
                df = pd.DataFrame({
                    # 'intensity_bin': np.array(list(hist.keys()), dtype=int),
                    'intensity_bin': np.array(list(hist.keys())),
                    'value': np.array(list(hist.values())),
                    'channel': [channel] * len(hist),
                    'sensor': [sensor] * len(hist),
                })
                to_stack[(channel, sensor)] = df

        full_df = pd.concat(list(to_stack.values()))
        is_finite = np.isfinite(full_df.intensity_bin)
        num_nonfinite = (~is_finite).sum()
        if num_nonfinite > 0:
            import warnings
            warnings.warn('There were {} non-finite values'.format(num_nonfinite))
            full_df = full_df[is_finite]
        full_df = full_df.reset_index()
        return full_df


@profile
def main(cmdline=True, **kwargs):
    r"""
    CommandLine:
        XDEV_PROFILE=1 xdoctest -m watch.cli.coco_spectra main

    Example:
        >>> from watch.cli.coco_spectra import *  # NOQA
        >>> import kwcoco
        >>> test_dpath = ub.Path.appdir('watch/tests/cli/spectra').ensuredir()
        >>> image_fpath = test_dpath + '/intensityhist_demo.jpg'
        >>> coco_dset = kwcoco.CocoDataset.demo('vidshapes-msi-multisensor-videos1-frames2-gsize8')
        >>> kwargs = {'src': coco_dset, 'dst': image_fpath, 'mode': 'thread'}
        >>> kwargs['multiple'] = 'layer'
        >>> kwargs['element'] = 'step'
        >>> kwargs['workers'] = 'avail'
        >>> kwargs['show'] = False
        >>> kwargs['draw'] = False
        >>> main(cmdline=False, **kwargs)

    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> from watch.cli.coco_spectra import *  # NOQA
        >>> import kwcoco
        >>> import watch
        >>> test_dpath = ub.Path.appdir('watch/tests/cli/spectra').ensuredir()
        >>> image_fpath = test_dpath + '/intensityhist_demo2.jpg'
        >>> coco_dset = watch.coerce_kwcoco('watch-msi')
        >>> kwargs = {
        >>>     'src': coco_dset,
        >>>     'dst': image_fpath,
        >>>     'mode': 'thread',
        >>>     'valid_range': '10:2000',
        >>>     'workers': 'avail',
        >>> }
        >>> kwargs['multiple'] = 'layer'
        >>> kwargs['element'] = 'step'
        >>> main(cmdline=False, **kwargs)
    """
    from watch.utils import kwcoco_extensions
    from watch.utils import util_parallel
    import watch
    import kwcoco
    import numpy as np

    config = CocoSpectraConfig.cli(data=kwargs, cmdline=cmdline)
    print('config = {}'.format(ub.urepr(config.to_dict(), nl=1)))

    # coco_dset = kwcoco.CocoDataset.coerce(config['src'])
    coco_dset = watch.coerce_kwcoco(config['src'])

    valid_gids = kwcoco_extensions.filter_image_ids(
        coco_dset,
        include_sensors=config['include_sensors'],
        exclude_sensors=config['exclude_sensors'],
        select_images=config['select_images'],
        select_videos=config['select_videos'],
    )

    images = coco_dset.images(valid_gids)
    workers = util_parallel.coerce_num_workers(config['workers'])
    print('workers = {!r}'.format(workers))

    if config['max_images'] is not None:
        print('images = {!r}'.format(images))
        images = coco_dset.images(list(images)[:int(config['max_images'])])
        print('filter images = {!r}'.format(images))

    include_channels = config['include_channels']
    exclude_channels = config['exclude_channels']
    include_channels = None if include_channels is None else kwcoco.FusedChannelSpec.coerce(include_channels)
    exclude_channels = None if exclude_channels is None else kwcoco.FusedChannelSpec.coerce(exclude_channels)

    jobs = ub.JobPool(mode=config['mode'], max_workers=workers, transient=True)
    from watch.utils import util_progress
    pman = util_progress.ProgressManager()
    with pman:
        for coco_img in pman.progiter(images.coco_images, desc='submit stats jobs'):
            coco_img.detach()
            job = jobs.submit(ensure_intensity_stats, coco_img,
                              include_channels=include_channels,
                              exclude_channels=exclude_channels)
            job.coco_img = coco_img

        if config['valid_range'] is not None:
            valid_min, valid_max = map(float, config['valid_range'].split(':'))
        else:
            valid_min = -math.inf
            valid_max = math.inf

        accum = HistAccum()
        seen_props = ub.ddict(set)
        for job in pman.progiter(jobs.as_completed(), total=len(jobs), desc='accumulate stats'):
            intensity_stats = job.result()
            sensor = job.coco_img.get('sensor_coarse', job.coco_img.get('sensor', 'unknown_sensor'))
            for band_stats in intensity_stats['bands']:
                intensity_hist = band_stats['intensity_hist']
                band_props = band_stats['properties']
                for k, v in band_props.items():
                    seen_props[k].add(v)
                band_name = band_props['band_name']
                intensity_hist = {k: v for k, v in intensity_hist.items()
                                  if k >= valid_min and k <= valid_max}
                accum.update(intensity_hist, sensor, band_name)
            pman.update_info(
                ('seen_props = {}'.format(ub.urepr(seen_props, nl=1)))
            )

        full_df = accum.finalize()

        sensor_chan_stats, distance_metrics = sensor_stats_tables(full_df)

        COMPARSE_SENSORS = True
        if COMPARSE_SENSORS:
            request_columns = ['emd', 'energy_dist', 'mean_diff', 'std_diff']
            have_columns = list(ub.oset(request_columns) & ub.oset(distance_metrics.columns))
            harmony_scores = distance_metrics[have_columns].mean()
            extra_text = ub.urepr(harmony_scores.to_dict(), precision=4, compact=1)
            print('extra_text = {!r}'.format(extra_text))
        else:
            extra_text = None

    if config['draw']:
        import kwplot
        kwplot.autosns()
        fig = plot_intensity_histograms(full_df, config)

        title_lines = []
        title = config.get('title', None)
        if title is not None:
            title_lines.append(title)

        title_lines.append(ub.Path(coco_dset.fpath).name)
        if extra_text is not None:
            title_lines.append(extra_text)

        final_title = '\n'.join(title_lines)
        fig.suptitle(final_title)

        dst_fpath = config['dst']
        if dst_fpath is not None:
            print('dump to dst_fpath = {!r}'.format(dst_fpath))
            fig.set_size_inches(np.array([6.4, 4.8]) * 1.68)
            fig.tight_layout()
            fig.savefig(dst_fpath)

        if config['show']:
            from matplotlib import pyplot as plt
            plt.show()

    results = {
        'sensor_chan_stats': sensor_chan_stats,
        'distance_metrics': distance_metrics,
    }
    return results


# def try_with_statsmodels():
#     import statsmodels.api as sm
#     sm.nonparametric.KDEUnivariate?
#     kde = sm.nonparametric.KDEUnivariate(obs_dist)


def sensor_stats_tables(full_df):
    import itertools as it
    import scipy
    import kwarray
    import scipy.stats
    import pandas as pd
    import numpy as np
    sensor_channel_to_vwf = {}
    for _sensor, sensor_df in full_df.groupby('sensor'):
        for channel, chan_df in sensor_df.groupby('channel'):
            _values = chan_df['intensity_bin']
            _weights = chan_df['value']
            norm_weights = _weights / _weights.sum()
            sensor_channel_to_vwf[(_sensor, channel)] = {
                'raw_values': _values,
                'raw_weights': _weights,
                'norm_weights': norm_weights,
                'sensorchan_df': chan_df,
            }

    print('compute per-sensor stats')
    print(ub.urepr(list(sensor_channel_to_vwf.keys()), nl=1))
    chan_to_group = ub.group_items(
        sensor_channel_to_vwf.keys(),
        [t[1] for t in sensor_channel_to_vwf.keys()]
    )
    chan_to_combos = {
        chan: list(it.combinations(group, 2)) for chan, group in chan_to_group.items()
    }
    to_compare = list(ub.flatten(chan_to_combos.values()))

    single_rows = []
    for sensor, channel in ub.ProgIter(sensor_channel_to_vwf, desc='compute stats'):
        sensor_data = sensor_channel_to_vwf[(sensor, channel)]
        values = sensor_data['raw_values']
        weights = sensor_data['raw_weights']
        sensorchan_df = sensor_data['sensorchan_df']

        # Note: the calculation of the variance depends on the type of
        # weighting we choose
        average = np.average(values, weights=weights)
        variance = np.average((values - average) ** 2, weights=weights)
        variance = variance * sum(weights) / (sum(weights) - 1)
        stddev = np.sqrt(variance)

        pytype = float if values.values.dtype.kind == 'f' else int
        auto_bins = _weighted_auto_bins(
            sensorchan_df, 'intensity_bin', 'value')

        info = {
            'min': pytype(values.min()),
            'max': pytype(values.max()),
            'mean': average,
            'std': stddev,
            'total_weight': weights.sum(),
            'channel': channel,
            'sensor': sensor,
            'auto_bins': auto_bins,
        }
        assert info['max'] >= info['min']
        # print('info = {}'.format(ub.urepr(info, nl=1, sort=False)))
        single_rows.append(info)

    sensor_chan_stats = pd.DataFrame(single_rows)
    sensor_chan_stats = sensor_chan_stats.set_index(['sensor', 'channel'])
    print(sensor_chan_stats)

    print('compare channels between sensors')
    pairwise_rows = []
    for item1, item2 in ub.ProgIter(to_compare, desc='comparse_sensors', verbose=1):
        sensor1, channel1 = item1
        sensor2, channel2 = item2
        assert channel1 == channel2
        channel = channel1

        row = {
            'sensor1': sensor1,
            'sensor2': sensor2,
            'channel': channel,
        }

        u_sensor_data = sensor_channel_to_vwf[(sensor1, channel1)]
        v_sensor_data = sensor_channel_to_vwf[(sensor2, channel2)]

        u_values, u_weights = ub.take(u_sensor_data, ['raw_values', 'raw_weights'])
        v_values, v_weights = ub.take(v_sensor_data, ['raw_values', 'raw_weights'])

        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wasserstein_distance.html#scipy.stats.wasserstein_distance
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.energy_distance.html#scipy.stats.energy_distance
        dist_inputs = dict(
            u_values=u_values, v_values=v_values, u_weights=u_weights,
            v_weights=v_weights)

        if 1:
            # TODO: normalize data such that density (or probability?) is 1.0
            row['emd'] = scipy.stats.wasserstein_distance(**dist_inputs)

        if 1:
            row['energy_dist'] = scipy.stats.energy_distance(**dist_inputs)

        u_stats = sensor_chan_stats.loc[(sensor1, channel1)]
        v_stats = sensor_chan_stats.loc[(sensor2, channel2)]
        row['mean_diff'] = abs(u_stats['mean'] - v_stats['mean'])
        row['std_diff'] = abs(u_stats['std'] - v_stats['std'])

        # TODO: robust alignment of pdfs
        if 1:
            cuv_values = np.union1d(u_values, v_values)
            cuv_values.sort()
            cu_weights = np.zeros(cuv_values.shape, dtype=np.float64)
            cv_weights = np.zeros(cuv_values.shape, dtype=np.float64)
            cu_indexes = np.where(kwarray.isect_flags(cuv_values, u_values))[0]
            cv_indexes = np.where(kwarray.isect_flags(cuv_values, v_values))[0]
            cu_weights[cu_indexes] = u_weights
            cv_weights[cv_indexes] = v_weights

            cu_weights * cv_weights

            row['kld'] = scipy.stats.entropy(cu_weights, cv_weights)

        pairwise_rows.append(row)

    distance_metrics = pd.DataFrame(pairwise_rows)
    print(distance_metrics.to_string())
    return sensor_chan_stats, distance_metrics


@profile
def ensure_intensity_sidecar(fpath, recompute=False):
    """
    Write statistics next to the image

    Example:
        >>> from watch.cli.coco_spectra import *  # NOQA
        >>> import kwimage
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('watch/tests/intensity_sidecar').ensuredir()
        >>> dpath.delete().ensuredir()
        >>> img = kwimage.grab_test_image(dsize=(16, 16))
        >>> img01 = kwimage.ensure_float01(img)
        >>> img255 = kwimage.ensure_uint255(img)
        >>> fpath1 = dpath / 'img01.tif'
        >>> fpath2 = dpath / 'img255.tif'
        >>> kwimage.imwrite(fpath1, img01)
        >>> kwimage.imwrite(fpath2, img255)
        >>> fpath = fpath1
        >>> stats_fpath1 = ensure_intensity_sidecar(fpath1)
        >>> fpath = fpath2
        >>> stats_fpath2 = ensure_intensity_sidecar(fpath1)
        >>> import pickle
        >>> pickle.loads(stats_fpath1.read_bytes())
        >>> pickle.loads(stats_fpath2.read_bytes())
    """
    import os
    import kwarray
    import kwimage
    import pickle
    stats_fpath = ub.Path(os.fspath(fpath) + '.stats_v1.pkl')
    if recompute or not stats_fpath.exists():
        imdata = kwimage.imread(fpath, backend='gdal', nodata_method='ma')
        imdata = kwarray.atleast_nd(imdata, 3)
        # TODO: even better float handling
        if imdata.dtype.kind == 'f':
            precision = 2
            # Round floats to p decimals to keep things semi-managable
            # binning floats is necessary due to fundamental limitations
            # We may need to decrease our resolution even further.
            imdata = imdata.round(precision)
        stats_info = {'bands': []}
        for imband in imdata.transpose(2, 0, 1):
            masked_data = imband.ravel()
            # Remove masked data
            data = masked_data.data[~masked_data.mask]
            intensity_hist = ub.dict_hist(data)
            intensity_hist = ub.sorted_keys(intensity_hist)
            stats_info['bands'].append({
                'intensity_hist': intensity_hist,
            })
        import safer
        with safer.open(stats_fpath, 'wb') as file:
            pickle.dump(stats_info, file)
    return stats_fpath


@profile
def ensure_intensity_stats(coco_img, recompute=False, include_channels=None, exclude_channels=None):
    """
    Ensures a sidecar file exists for the kwcoco image
    """
    from os.path import join
    import numpy as np
    import kwcoco
    import kwimage
    import pickle
    intensity_stats = {'bands': []}
    for obj in coco_img.iter_asset_objs():
        fpath = join(coco_img.bundle_dpath, obj['file_name'])
        channels = obj.get('channels', None)
        if channels is None:
            shape = kwimage.load_image_shape(fpath)
            num_channels = shape[2]
            if num_channels == 3:
                channels = 'r|g|b'
            elif num_channels == 4:
                channels = 'r|g|b|a'
            elif num_channels == 1:
                channels = 'r|g|b'
            else:
                channels = kwcoco.FusedChannelSpec.coerce(num_channels)

        channels = kwcoco.FusedChannelSpec.coerce(channels)
        declared_channel_list = channels.as_list()

        quantization = obj.get('quantization', None)

        requested_channels = channels
        if include_channels:
            requested_channels = requested_channels & include_channels
        if exclude_channels:
            requested_channels = requested_channels - exclude_channels

        if requested_channels.numel() > 0:
            stats_fpath = ensure_intensity_sidecar(fpath, recompute=recompute)
            try:
                with open(stats_fpath, 'rb') as file:
                    stat_info = pickle.load(file)
            except Exception as ex:
                print('ex = {!r}'.format(ex))
                print('error loading stats_fpath = {!r}'.format(stats_fpath))
                raise

            alwaysappend = requested_channels.numel() == channels.numel()

            for band_idx, band_stat in enumerate(stat_info['bands']):
                try:
                    band_name = declared_channel_list[band_idx]
                except IndexError:
                    print('bad channel declaration fpath = {!r}'.format(fpath))
                    if 0:
                        print('obj = {}'.format(ub.urepr(obj, nl=1)))
                        print('coco_img = {!r}'.format(coco_img))
                        print('fpath = {!r}'.format(fpath))
                        print('stats_fpath = {!r}'.format(stats_fpath))
                        print(len(stat_info['bands']))
                        print('band_idx = {!r}'.format(band_idx))
                        print('channels = {!r}'.format(channels))
                    # raise
                    band_name = 'unknown'
                if quantization is not None:
                    # Handle dequantization
                    from delayed_image.helpers import dequantize
                    quant_values = np.array(list(band_stat['intensity_hist'].keys()))
                    counts = list(band_stat['intensity_hist'].values())
                    dequant_values = dequantize(quant_values, quantization)
                    band_stat['intensity_hist'] = ub.odict(zip(dequant_values, counts))
                if alwaysappend or band_name in requested_channels:
                    band_stat['properties'] = props = {
                        'video_name': None,
                        'band_name': band_name,
                        'month': None,
                    }
                    try:
                        props['video_name'] = coco_img.video['name']
                    except Exception:
                        ...
                    try:
                        from watch.utils import util_time
                        props['month'] = util_time.coerce_datetime(coco_img.img['date_captured']).month
                    except Exception:
                        ...
                    band_stat['band_name'] = band_name
                    intensity_stats['bands'].append(band_stat)
    return intensity_stats


@profile
def plot_intensity_histograms(full_df, config):
    unique_channels = full_df['channel'].unique()
    unique_sensors = full_df['sensor'].unique()

    import kwplot
    sns = kwplot.autosns()

    palette = {
        'red': 'red',
        'blue': 'blue',
        'green': 'green',
        'cirrus': 'skyblue',
        'coastal': 'purple',
        'nir': 'orange',
        'swir16': 'pink',
        'swir22': 'hotpink',
    }
    if 'red' not in unique_channels and 'r' in unique_channels:
        # hack
        palette['r'] = 'red'
        palette['g'] = 'green'
        palette['b'] = 'blue'
    for channel in unique_channels:
        if channel not in palette:
            palette[channel] = None
    palette = _fill_missing_colors(palette)

    hist_data_kw = dict(
        x='intensity_bin',
        weights='value',
        bins=config['bins'],
        stat=config['stat'],
        hue='channel',
    )
    hist_style_kw = dict(
        palette=palette,
        fill=config['fill'],
        element=config['element'],
        multiple=config['multiple'],
        kde=config['kde'],
        cumulative=config['cumulative'],
    )

    if 0:
        # __hisplot_notes__
        import inspect
        sig = inspect.signature(sns.histplot)
        # Print params we might not have looked at in detail
        exposed_params = {
            'cumulative', 'kde', 'multiple', 'element', 'fill', 'hue', 'stat',
            'bins', 'weights', 'x', 'palette',
        }
        probably_ignorable_params = {
            'pmax', 'hue_order', 'hue_norm', 'cbar', 'cbar_kws', 'cbar_ax', 'ax',
            'legend', 'thresh' 'y',
        }
        maybe_expose = (set(sig.parameters) - exposed_params) - probably_ignorable_params
        print('maybe_expose = {}'.format(ub.urepr(maybe_expose, nl=1)))

    #  For S2 that is supposed to be divide by 10000.  For L8 it is multiply by 2.75e-5 and subtract 0.2.
    # 1 / 2.75e-5
    print('start plot')
    fig = kwplot.figure(fnum=1, doclf=True)
    print('fig = {!r}'.format(fig))
    pnum_ = kwplot.PlotNums(nSubplots=len(unique_sensors))
    print('pnum_ = {!r}'.format(pnum_))
    for sensor_name, sensor_df in full_df.groupby('sensor'):
        print('plot sensor_name = {!r}'.format(sensor_name))

        hist_data_kw_ = hist_data_kw.copy()
        if hist_data_kw_['bins'] == 'auto':
            xvar = hist_data_kw['x']
            weightvar = hist_data_kw['weights']
            hist_data_kw_['bins'] = _weighted_auto_bins(sensor_df, xvar, weightvar)

        ax = kwplot.figure(fnum=1, pnum=pnum_()).gca()
        # z = [tuple(a.values()) for a in sensor_df[['intensity_bin', 'channel', 'sensor']].to_dict('records')]
        # ub.find_duplicates(z)
        try:
            # https://github.com/mwaskom/seaborn/issues/2709
            sns.histplot(ax=ax, data=sensor_df.reset_index(), **hist_data_kw_, **hist_style_kw)
        except Exception:
            print('hist_data_kw_ = {}'.format(ub.urepr(hist_data_kw_, nl=1)))
            print('hist_style_kw = {}'.format(ub.urepr(hist_style_kw, nl=1)))
            print('ERROR')
            print(sensor_df)
            raise
            pass
        if config['valid_range'] is not None:
            valid_min, valid_max = map(float, config['valid_range'].split(':'))
            ax.set_xlim(valid_min, valid_max)

        ax.set_title(sensor_name)
        # maxx = sensor_df.intensity_bin.max()
        # maxx = sensor_maxes[sensor_name]
        # ax.set_xlim(0, maxx)

    return fig


# def _weighted_quantile(weights, qs):
#     cumtotal = np.cumsum(weights)
#     quantiles = cumtotal / cumtotal[-1]


@profile
def _weighted_auto_bins(data, xvar, weightvar):
    """
    Generalized histogram bandwidth estimators for weighted univariate data

    References:
        https://github.com/mwaskom/seaborn/issues/2710

    TODO:
        add to util_kwarray

    Example:
        >>> import pandas as pd
        >>> import numpy as np
        >>> n = 100
        >>> to_stack = []
        >>> rng = np.random.RandomState(432)
        >>> for group_idx in range(3):
        >>>     part_data = pd.DataFrame({
        >>>         'x': np.arange(n),
        >>>         'weights': rng.randint(0, 100, size=n),
        >>>         'hue': [f'group_{group_idx}'] * n,
        >>>     })
        >>>     to_stack.append(part_data)
        >>> data = pd.concat(to_stack).reset_index()
        >>> xvar = 'x'
        >>> weightvar = 'weights'
        >>> n_equal_bins = _weighted_auto_bins(data, xvar, weightvar)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import seaborn as sns
        >>> sns.histplot(data=data, bins=n_equal_bins, x='x', weights='weights', hue='hue')
    """
    import numpy as np
    sort_df = data.sort_values(xvar)
    values = sort_df[xvar]
    weights = sort_df[weightvar]
    minval = values.iloc[0]
    maxval = values.iloc[-1]

    total = weights.sum()
    ptp = maxval - minval

    # _hist_bin_sqrt = ptp / np.sqrt(total)
    _hist_bin_sturges = ptp / (np.log2(total) + 1.0)

    cumtotal = weights.cumsum().values
    quantiles = cumtotal / cumtotal[-1]
    idx2, idx1 = np.searchsorted(quantiles, [0.75, 0.25])
    # idx2, idx1 = _weighted_quantile(weights, [0.75, 0.25])
    iqr = values.iloc[idx2] - values.iloc[idx1]
    _hist_bin_fd = 2.0 * iqr * total ** (-1.0 / 3.0)

    fd_bw = _hist_bin_fd  # Freedman-Diaconis
    sturges_bw = _hist_bin_sturges

    if fd_bw:
        bw_est = min(fd_bw, sturges_bw)
    else:
        # limited variance, so we return a len dependent bw estimator
        bw_est = sturges_bw

    from numpy.lib.histograms import _get_outer_edges, _unsigned_subtract
    first_edge, last_edge = _get_outer_edges(values, None)
    if bw_est:
        n_equal_bins = int(np.ceil(_unsigned_subtract(last_edge, first_edge) / bw_est))
    else:
        # Width can be zero for some estimators, e.g. FD when
        # the IQR of the data is zero.
        n_equal_bins = 1

    # Take the minimum of this and the number of actual bins
    n_equal_bins = min(n_equal_bins, len(values))
    return n_equal_bins


@profile
def _fill_missing_colors(label_to_color):
    """
    label_to_color = {'foo': kwimage.Color('red').as01(), 'bar': None}
    """
    from distinctipy import distinctipy
    import kwarray
    import numpy as np
    import kwimage
    given = {k: kwimage.Color(v).as01() for k, v in label_to_color.items() if v is not None}
    needs_color = sorted(set(label_to_color) - set(given))

    seed = 6777939437
    # hack in our code

    def _patched_get_random_color(pastel_factor=0, rng=None):
        rng = kwarray.ensure_rng(seed, api='python')
        color = [(rng.random() + pastel_factor) / (1.0 + pastel_factor) for _ in range(3)]
        return tuple(color)
    distinctipy.get_random_color = _patched_get_random_color

    exclude_colors = [
        tuple(map(float, (d, d, d)))
        for d in np.linspace(0, 1, 5)
    ] + list(given.values())

    final = given.copy()
    new_colors = distinctipy.get_colors(len(needs_color), exclude_colors=exclude_colors)
    for key, new_color in zip(needs_color, new_colors):
        final[key] = tuple(map(float, new_color))
    return final


__config__ = CocoSpectraConfig

if __name__ == '__main__':
    main()



import argparse
import sys
import json
import os
import tempfile
import subprocess

import pystac

from watch.stac.util_stac import parallel_map_items


def main():
    parser = argparse.ArgumentParser(
        description="Egress data to T&E baseline framework structure")

    parser.add_argument('stac_catalog',
                        type=str,
                        help="Path to input STAC catalog")
    parser.add_argument('output_path',
                        type=str,
                        help="S3 path for output JSON")
    parser.add_argument("-o", "--outbucket",
                        type=str,
                        required=True,
                        help="S3 Output directory for STAC item / asset "
                             "egress")
    parser.add_argument("--aws_profile",
                        required=False,
                        type=str,
                        help="AWS Profile to use for AWS S3 CLI commands")
    parser.add_argument("--aws_storage_class",
                        required=False,
                        type=str,
                        default=None,
                        help="AWS S3 storage class to use for egress AWS S3 "
                             "CLI commands (e.g. 'ONEZONE_IA', default: None)")
    parser.add_argument("-d", "--dryrun",
                        action='store_true',
                        default=False,
                        help="Run AWS CLI commands with --dryrun flag")
    parser.add_argument('-s', '--show-progress',
                        action='store_true',
                        default=False,
                        help='Show progress for AWS CLI commands')
    parser.add_argument("-n", "--newline",
                        action='store_true',
                        default=False,
                        help="Output as simple newline separated STAC items")
    parser.add_argument("-j", "--jobs",
                        type=int,
                        default=1,
                        required=False,
                        help="Number of jobs to run in parallel")

    baseline_framework_egress(**vars(parser.parse_args()))

    return 0


def egress_item(stac_item, outbucket, aws_base_command):
    if isinstance(stac_item, dict):
        stac_item_dict = stac_item
    elif isinstance(stac_item, pystac.Item):
        stac_item_dict = stac_item.to_dict()
    else:
        raise TypeError("Expecting 'stac_item' to be either a dictionary "
                        "or pystac.Item")

    stac_item_outpath = os.path.join(
        outbucket, "{}.json".format(stac_item.id))

    assets_outdir = os.path.join(outbucket, stac_item.id)

    for asset_name, asset in stac_item_dict.get('assets', {}).items():
        asset_basename = os.path.basename(asset['href'])

        asset_outpath = os.path.join(assets_outdir, asset_basename)

        command = [*aws_base_command, asset['href'], asset_outpath]

        print("Running: {}".format(' '.join(command)))
        # TODO: Manually check return code / output
        subprocess.run(command, check=True)

        # Update feature asset href to point to local outpath
        asset['href'] = asset_outpath

    with tempfile.NamedTemporaryFile() as temporary_file:
        with open(temporary_file.name, 'w') as f:
            print(json.dumps(stac_item_dict, indent=2), file=f)

        command = [*aws_base_command,
                   temporary_file.name, stac_item_outpath]

        subprocess.run(command, check=True)

    output_stac_item = pystac.Item.from_dict(stac_item_dict)
    output_stac_item.set_self_href(stac_item_outpath)
    return output_stac_item


def upload_output_stac_items(output_stac_items,
                             output_path,
                             aws_base_command,
                             newline=False):
    if newline:
        te_output = '\n'.join((json.dumps(item) for item in output_stac_items))
    else:
        te_output = {'raw_images': [],
                     'stac': {
                         'type': 'FeatureCollection',
                         'features': output_stac_items}}

    with tempfile.NamedTemporaryFile() as temporary_file:
        with open(temporary_file.name, 'w') as f:
            if newline:
                print(te_output, file=f)
            else:
                print(json.dumps(te_output, indent=2), file=f)

        command = [*aws_base_command, temporary_file.name, output_path]

        subprocess.run(command, check=True)

    return te_output


def baseline_framework_egress(stac_catalog,
                              output_path,
                              outbucket,
                              aws_profile=None,
                              dryrun=False,
                              show_progress=False,
                              newline=False,
                              jobs=1,
                              aws_storage_class=None):
    if isinstance(stac_catalog, str):
        catalog = pystac.read_file(href=stac_catalog).full_copy()
    elif isinstance(stac_catalog, dict):
        catalog = pystac.Catalog.from_dict(stac_catalog).full_copy()
    else:
        catalog = stac_catalog.full_copy()

    if aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    if dryrun:
        aws_base_command.append('--dryrun')

    if not show_progress:
        aws_base_command.append('--no-progress')

    if aws_storage_class is not None:
        aws_base_command.extend(['--storage-class', aws_storage_class])

    output_catalog = parallel_map_items(
        catalog,
        egress_item,
        max_workers=jobs,
        mode='process' if jobs > 1 else 'serial',
        extra_args=[outbucket, aws_base_command])

    output_stac_items = [item.to_dict() for item
                         in output_catalog.get_all_items()]

    te_output = upload_output_stac_items(
        output_stac_items, output_path, aws_base_command, newline=newline)

    return te_output


if __name__ == "__main__":
    sys.exit(main())



"""
This submodule contains tools relating to:

    * the IARPA region and site model geojson formats.
    * geographic kwcoco extensions

At the time of creation we have many various tools scattered across the repo to
deal with these, but we should work to consolidate the ones pertaining strictly
to these specific types of geojsons into this submodule.


Location of existing tools that should likely be moved:

    * ../demo/metrics_demo/__init__.py :: various
    * ../cli/kwcoco_to_geojson.py :: various
    * ../cli/reproject_annotations.py :: various
    * ../cli/validate_annotation_schemas.py :: various
    * ../cli/extend_sc_sites.py :: various
    * ../cli/crop_sites_to_regions.py :: various
    * ../cli/cluster_sites.py :: various
    * ../utils/util_framework.py :: determine_region_id

    * ~/code/watch/dev/poc/make_region_from_sitemodel.py


Related tool that should NOT be moved are related to general geojson:

    * ../utils/util_gis.py :: coerce_geojson_datas
"""



"""
Geojson object oriented interface for region and site models

SeeAlso:
    ../rc/registry.py
"""
import ubelt as ub
import geopandas as gpd
import geojson
import jsonschema
import copy
from watch.utils import util_time
from watch.utils import util_progress


class _Model(ub.NiceRepr, geojson.FeatureCollection):
    type = 'FeatureCollection'
    _header_type = NotImplemented
    _body_type = NotImplemented

    def __nice__(self):
        return ub.urepr(self.info(), nl=2)

    def pandas(self):
        """
        Returns:
            geopandas.GeoDataFrame: the feature collection as data frame
        """
        gdf = gpd.GeoDataFrame.from_features(self['features'])
        return gdf

    def deepcopy(self):
        return copy.deepcopy(self)

    def dumps(self, **kw):
        import json
        return json.dumps(self, **kw)

    @classmethod
    def coerce_multiple(cls, data):
        from watch.utils import util_gis
        infos = list(util_gis.coerce_geojson_datas(data, format='json'))
        for info in infos:
            yield cls(**info['data'])

    @classmethod
    def coerce(cls, data):
        import os
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls.from_dict(data)
        elif isinstance(data, list):
            if all(isinstance(d, dict) and d['type'] == 'Feature' for d in data):
                return cls.from_features(data)
            else:
                raise TypeError('lists must a list of Features')
            return cls.from_dict(data)
        elif isinstance(data, gpd.GeoDataFrame):
            return cls.from_dataframe(data)
        elif isinstance(data, (str, os.PathLike)):
            got = list(cls.coerce_multiple(data))
            assert len(got) == 1
            return got[0]
        else:
            raise TypeError

    @classmethod
    def from_features(cls, features):
        """
        Args:
            gdf (GeoDataFrame):
        """
        self = cls(features=features)
        return self

    @classmethod
    def from_dataframe(cls, gdf):
        """
        Args:
            gdf (GeoDataFrame):
        """
        jsondct = gdf.__geo_interface__
        return cls(**jsondct)

    @classmethod
    def from_dict(cls, data):
        """
        Args:
            gdf (GeoDataFrame):
        """
        return cls(**data)

    @property
    def start_date(self):
        return util_time.coerce_datetime(self.header['properties']['start_date'])

    @property
    def end_date(self):
        return util_time.coerce_datetime(self.header['properties']['end_date'])

    def load_schema(self, strict=True):
        raise NotImplementedError('abstract')

    def body_features(self):
        for feat in self['features']:
            prop = feat['properties']
            if prop['type'] == self._body_type:
                yield feat

    @property
    def header(self):
        for feat in self['features']:
            prop = feat['properties']
            if prop['type'] == self._header_type:
                return feat

    def _validate_quick_checks(self):
        header = self.header
        if header is None:
            raise AssertionError('Geo Model has no header')

        if header is not self.features[0]:
            raise AssertionError('Header should be the first feature')

        if header['properties']['type'] != self._header_type:
            raise AssertionError('Header type is wrong')

        if self['type'] != 'FeatureCollection':
            raise AssertionError('GeoModels should be FeatureCollections')

        feature_types = ub.dict_hist([
            f['properties']['type'] for f in self.features])
        assert feature_types.pop(self._header_type, 0) == 1
        assert set(feature_types).issubset({self._body_type})

        start_date = self.start_date
        end_date = self.end_date
        if start_date is not None and end_date is not None:
            if end_date < start_date:
                raise AssertionError('bad date')

    def _validate_schema(self, strict=True):
        import rich

        def print_validation_error_info(ex, depth=1):
            if ex.parent is not None:
                max_depth = print_validation_error_info(ex.parent, depth=depth + 1)
            else:
                max_depth = depth
            rich.print(f'[yellow] error depth = {depth} / {max_depth}')
            print('ex.__dict__ = {}'.format(ub.urepr(ex.__dict__, nl=3)))
            return depth

        schema = self.load_schema(strict=strict)
        try:
            jsonschema.validate(self, schema)
        except jsonschema.ValidationError as e:
            ex = e
            rich.print('[red] JSON VALIDATION ERROR')
            print(f'self={self}')
            print_validation_error_info(ex)
            # ub.IndexableWalker(self)[ex.absolute_path]
            # ub.IndexableWalker(schema)[ex.schema_path]
            rich.print(ub.codeblock(
                '''
                [yellow] jsonschema validation notes:
                    * depsite our efforts, information to debug the issue may not be shown, double check your schema and instance manually.
                    * anyOf schemas may print the error, and not the part you intended to match.
                    * oneOf schemas may not explicitly say that you matched both.
                '''))
            rich.print('[red] JSON VALIDATION ERROR')
            raise

    def validate(self, strict=True):
        self._validate_quick_checks()
        self._validate_schema(strict=strict)


class RegionModel(_Model):
    """
    Wrapper around a geojson region model FeatureCollection

    Example:
        >>> from watch.geoannots.geomodels import *  # NOQA
        >>> self = RegionModel.random()
        >>> print(self)
        >>> self.validate(strict=False)

    """
    _header_type = 'region'
    _body_type = 'site_summary'

    def info(self):
        header = self.header
        prop = '<no region header>' if header is None else header['properties']
        info = {
            'num_site_summaries': len(list(self.site_summaries())),
            'properties': prop,
        }
        return info

    def load_schema(self, strict=True):
        import watch
        schema = watch.rc.registry.load_region_model_schema(strict=strict)
        return schema

    def site_summaries(self):
        yield from (SiteSummary(**f) for f in self.body_features())

    @classmethod
    def coerce(cls, data):
        """
        Example:
            >>> from watch.geoannots.geomodels import *  # NOQA
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('watch/tests/geoannots/coerce').ensuredir()
            >>> region = RegionModel.random(with_sites=False, rng=0)
            >>> data = fpath = (dpath/ 'region.geojson')
            >>> fpath.write_text(region.dumps())
            >>> region_models = list(RegionModel.coerce_multiple(fpath))
            >>> region_model = RegionModel.coerce(fpath)
        """
        self = super().coerce(data)
        assert self.header['properties']['type'] == 'region'
        return self

    def pandas_summaries(self):
        """
        Returns:
            geopandas.GeoDataFrame: the site summaries as a data frame
        """
        gdf = gpd.GeoDataFrame.from_features(list(self.site_summaries()))
        return gdf

    def pandas_region(self):
        """
        Returns:
            geopandas.GeoDataFrame: the region header as a data frame
        """
        gdf = gpd.GeoDataFrame.from_features([self.header])
        return gdf

    @classmethod
    def random(cls, with_sites=False, **kwargs):
        """
        Args:
            with_sites (bool):
                also returns site models if True
            **kwargs : passed to :func:`demo_truth.random_region_model`

        Returns:
            RegionModel | Tuple[RegionModel, SiteModelCollection]

        Example:
            >>> from watch.geoannots.geomodels import *  # NOQA
            >>> region1 = RegionModel.random(with_sites=False, rng=0)
            >>> region2, sites2 = RegionModel.random(with_sites=True, rng=0)
            >>> assert region1 == region2, 'rngs should be the same'
        """
        from watch.demo.metrics_demo import demo_truth

        region, sites, _ = demo_truth.random_region_model(
            **kwargs, with_renderables=False)

        region = cls(**region)

        if with_sites:
            sites = SiteModelCollection([SiteModel(**s) for s in sites])
            return region, sites
        else:
            return region

    @property
    def region_id(self):
        return self.header['properties']['region_id']


class SiteModel(_Model):
    """
    Wrapper around a geojson site model FeatureCollection

    Example:
        >>> from watch.geoannots.geomodels import *  # NOQA
        >>> self = SiteModel.random()
        >>> print(self)
        >>> self.validate(strict=False)
    """
    _header_type = 'site'
    _body_type = 'observation'

    def info(self):
        header = self.header
        prop = '<no site header>' if header is None else header['properties']
        info = {
            'num_observations': len(list(self.observations())),
            'properties': prop,
        }
        return info

    def load_schema(self, strict=True):
        import watch
        schema = watch.rc.registry.load_site_model_schema(strict=strict)
        return schema

    @property
    def header(self):
        for feat in self['features']:
            prop = feat['properties']
            if prop['type'] == 'site':
                return feat

    def observations(self):
        yield from self.body_features()

    def pandas_observations(self):
        """
        Returns:
            geopandas.GeoDataFrame: the site summaries as a data frame
        """
        gdf = gpd.GeoDataFrame.from_features(list(self.observations()))
        return gdf

    @classmethod
    def random(cls, rng=None, **kwargs):
        """
        """
        from watch.demo.metrics_demo import demo_truth
        _, sites, _ = demo_truth.random_region_model(num_sites=1, rng=rng, **kwargs)
        return cls(**sites[0])

    def as_summary(self):
        header = self.header
        if header is None:
            ...
        else:
            summary = header.copy()
            summary['properties'] = summary['properties'].copy()
            assert summary['properties']['type'] == 'site'
            summary['properties']['type'] = 'site_summary'
            return SiteSummary(**summary)

    @property
    def site_id(self):
        return self.header['properties']['site_id']

    @property
    def status(self):
        return self.header['properties']['status']

    def fix_geom(self):
        from shapely.geometry import shape
        from shapely.validation import make_valid
        from shapely.geometry import MultiPolygon
        for feat in self.features:
            geom = shape(feat['geometry'])
            if geom.geom_type in {'MultiPolygon', 'Polygon'}:
                make_valid(geom)
            else:
                geom = geom.buffer(3).convex_hull
                geom = MultiPolygon([geom])
            feat['geometry'] = geom.__geo_interface__

    def fixup(self):
        self.clamp_scores()
        # self.fix_geom()

    def clamp_scores(self):
        for feat in self.features:
            fprop = feat['properties']
            fprop['score'] = float(max(min(1, fprop['score']), 0))


class _Feature(ub.NiceRepr, geojson.Feature):
    type = 'Feature'

    def __nice__(self):
        return ub.urepr(self.info(), nl=2)

    def info(self):
        info = {
            'properties': self['properties'],
        }
        return info


class Observation(_Feature):
    ...


class _SiteOrSummary(_Feature):
    """
    Site summaries and site headers are nearly the same
    """

    @property
    def start_date(self):
        return util_time.coerce_datetime(self['properties']['start_date'])

    @property
    def end_date(self):
        return util_time.coerce_datetime(self['properties']['end_date'])

    @property
    def site_id(self):
        return self['properties']['site_id']


class SiteSummary(_SiteOrSummary):
    ...


class SiteHeader(_SiteOrSummary):
    ...


class RegionHeader(_Feature):
    """
    A helper wrapper for the region model header features
    """

    @classmethod
    def coerce(cls, data):
        """
        Example:
            >>> data = RegionModel.random()
            >>> h1 = RegionHeader.coerce(data)
            >>> h2 = RegionHeader.coerce(data.header)
            >>> assert h1 == h2

            RegionHeader.coerce(orig_region_model)
        """
        if isinstance(data, cls):
            return data
        if isinstance(data, dict):
            if 'type' in data:
                if data['type'] == 'Feature':
                    return cls(**data)
                elif data['type'] == 'FeatureCollection':
                    return cls(**RegionModel(**data).header)
        raise TypeError(data)


# def _site_header_from_observations(observations, mgrs_code, site_id, status, summary_geom=None):
#     """
#     Consolodate site observations into a site summary
#     """
#     if summary_geom is None:
#         summary_geom = unary_union(
#             [kwimage.MultiPolygon.coerce(o["geometry"]).to_shapely() for o in observations]
#         ).convex_hull
#     start_date = observations[0]["properties"]["observation_date"]
#     end_date = observations[-1]["properties"]["observation_date"]
#     sitesum_props = {
#         "type": "site_summary",
#         "status": status,
#         "version": "2.0.1",
#         "site_id": site_id,
#         "mgrs": mgrs_code,
#         "start_date": start_date,
#         "end_date": end_date,
#         "score": 1,
#         "originator": "demo",
#         "model_content": "annotation",
#         "validated": "True",
#     }
#     site_summary = geojson.Feature(
#         properties=sitesum_props,
#         geometry=kwimage.Polygon.coerce(summary_geom).to_geojson(),
#     )
#     return site_summary


class ModelCollection(list):
    """
    A storage container for multiple site / region models
    """

    def fixup(self):
        pman = util_progress.ProgressManager()
        with pman:
            for s in pman.progiter(self, desc='fixup'):
                s.fixup()

    def validate(self, mode='process', workers=0):
        import rich
        # pman = util_progress.ProgressManager(backend='progiter')
        pman = util_progress.ProgressManager()
        with pman:
            tries = 0
            while True:
                jobs = ub.JobPool(mode='process', max_workers=8 if tries == 0 else 0)
                for s in pman.progiter(self, desc='submit validate models'):
                    job = jobs.submit(s.validate)
                    job.s = s
                try:
                    for job in pman.progiter(jobs.as_completed(), total=len(jobs), desc='collect validate models'):
                        job.result()
                except Exception:
                    if tries > 0 or workers == 0:
                        raise Exception
                    tries += 1
                    rich.print('[red] ERROR: [yellow] Failed to validate: trying again with workers=0')
                else:
                    break


class SiteModelCollection(ModelCollection):

    def as_region_model(self, region=None):
        """
        Convert a set of site models to a region model

        Args:
            region (RegonModel | RegionHeader):
                updates an existing region model with new site summaries
        """
        site_summaries = [s.as_summary() for s in self]

        if region is not None:
            region_header = RegionHeader.coerce(region)
        else:
            raise NotImplementedError
            region_header = geojson.Feature(
                properties={
                    "type": "region",
                    "region_id": None,
                    "version": "2.4.3",
                    "mgrs": None,
                    "start_date": None,
                    "end_date": None,
                    "originator": None,
                    "model_content": None,
                    "comments": "",
                },
                geometry=None,
            )

        region_geom = region_header['geometry']
        if region_geom is None:
            from shapely.ops import unary_union
            import shapely.geometry
            site_geoms = [shapely.geometry.shape(s['geometry']).buffer(0)
                          for s in site_summaries]
            region_geom = unary_union(site_geoms).convex_hull

        region_features = [region_header] + site_summaries
        region_model = RegionModel(features=region_features)
        return region_model



import torch
import kwcoco
import random
import numpy as np

import albumentations as A


class kwcoco_dataset(torch.utils.data.Dataset):
    S2_channel_names = [
        'coastal', 'blue', 'green', 'red', 'B05', 'B06', 'B07', 'nir', 'B09', 'cirrus', 'swir16', 'swir22', 'B8A'
    ]
    L8_channel_names = [
        'coastal', 'lwir11', 'lwir12', 'blue', 'green', 'red', 'nir', 'swir16', 'swir22', 'pan', 'cirrus'
    ]
    common_channel_names = [
        'coastal', 'blue', 'green', 'red', 'nir', 'swir16', 'swir22', 'cirrus'
    ]

    def __init__(self, coco_dset, sensor='S2', bands='all', patch_size=64, mode='train'):
        # initialize dataset
        self.dset = kwcoco.CocoDataset.coerce(coco_dset)
        self.images = self.dset.images()

        # handle if there are multiple sensors
        if sensor != 'all':
            if 'sensor_coarse' in self.images._id_to_obj[self.images._ids[0]].keys():
                # get available sensors
                avail_sensors = set(self.images.lookup('sensor_coarse'))
                # invalid sensor
                if sensor is None or sensor not in avail_sensors:
                    raise ValueError(f'sensor must be specified. Options are {", ".join(avail_sensors)}')
                # filter images by desired sensor
                else:
                    self.images = self.images.compress([x == sensor for x in self.images.lookup('sensor_coarse')])
                    assert self.images
            else:
                avail_sensors = None

        # get image ids and videos
        self.dset_ids = self.images.gids
        self.videos = [x['id'] for x in self.dset.videos().objs]

        # get all available channels
        all_channels = [ aux.get('channels', None) for aux in self.dset.index.imgs[self.images._ids[0]].get('auxiliary', []) ]
        if 'r|g|b' in all_channels:
            all_channels.remove('r|g|b')
        self.channels = []
        # no channels selected

        if bands == 'all':
            bands = '|'.join(all_channels)
        elif bands == 'S2':
            bands = '|'.join(self.S2_channel_names)
        elif bands == 'L8':
            bands = '|'.join(self.L8_channel_names)
        elif bands == 'common':
            bands = '|'.join(self.L8_channel_names)

        self.channels = kwcoco.FusedChannelSpec.coerce(bands)
        # if len(bands) < 1:
        #     raise ValueError(f'bands must be specified. Options are {", ".join(all_channels)}, or all')
        # # all channels selected
        # elif len(bands) == 1 and bands[0].lower() == 'all':
        #     self.channels = all_channels
        # # subset of channels selected
        # else:
        #     for band in bands:
        #         if band in all_channels:
        #             self.channels.append(band)
        #         else:
        #             raise ValueError(f'\'{band}\' not recognized as an available band. Options are {", ".join(all_channels)}, or all')

        # define augmentations
        self.transforms = A.Compose([
                A.RandomCrop(height=patch_size, width=patch_size),
                A.RandomRotate90(p=.999)],
            additional_targets={'image2': 'image'})

        self.transforms2 = A.Compose([
                A.RandomCrop(height=patch_size, width=patch_size),
                A.RandomRotate90(p=.999),
                A.HorizontalFlip(p=0.999)],
            additional_targets={'image2': 'image'})

        self.transforms3 = A.Compose([
                A.Blur(p=.75),
                A.RandomBrightnessContrast(brightness_by_max=False, always_apply=True)
        ])

        self.mode = mode

    def __len__(self,):
        return len(self.dset_ids)

    def get_img(self, idx, device=None):
        image_id = self.dset_ids[idx]
        image_info = self.dset.index.imgs[image_id]
        delayed = self.dset.delayed_load(
            image_id, channels=self.channels, space='video')
        image = delayed.finalize()
        real_mean = np.nanmean(image)
        real_std = np.nanstd(image)
        image = np.nan_to_num(image)
        image = image.astype(np.float32)
        image = torch.tensor(image)
        # normalize
        if real_std != 0.0:
            image = (image - real_mean) / real_std
        if device:
            image = image.to(device)
        image = image.permute(2, 0, 1).unsqueeze(0)
        return image_id, image_info, image

    def __getitem__(self, idx):
        if self.mode == 'test':
            # hack
            return self.get_img(idx)

        # get image1 id and the video it is associated with
        img1_id = self.dset_ids[idx]
        video = [ y for y in self.videos if img1_id in self.dset.index.vidid_to_gids[y] ][0]

        # randomly select image2 id from the same video (could be before or after image1)
        # make sure image2 is not image1 and image2 is in the set of filtered images by desired sensor
        img2_id = img1_id
        while img2_id == img1_id or img2_id not in self.dset_ids:
            img2_id = random.choice(self.dset.index.vidid_to_gids[video])

        # get frame indices for each image (used to determine which image was captured first)
        frame_index1 = self.dset.index.imgs[img1_id]['frame_index']
        frame_index2 = self.dset.index.imgs[img2_id]['frame_index']

        # load images
        img1 = self.dset.delayed_load(img1_id, channels=self.channels, space='video').finalize().astype(np.float32)
        img2 = self.dset.delayed_load(img2_id, channels=self.channels, space='video').finalize().astype(np.float32)
        img1 = np.nan_to_num(img1)
        img2 = np.nan_to_num(img2)

        # normalize
        if img1.std() != 0.0:
            img1 = (img1 - img1.mean()) / img1.std()
        if img2.std() != 0.0:
            img2 = (img2 - img2.mean()) / img2.std()

        # transformations
        transformed = self.transforms(image=img1, image2=img2)
        transformed2 = self.transforms2(image=img1)

        img1 = transformed['image']
        img2 = transformed['image2']
        img3 = transformed2['image']

        transformed3 = self.transforms3(image=img1)
        img4 = transformed3['image']

        # convert to tensors
        img1 = torch.tensor(img1).permute(2, 0, 1)
        img2 = torch.tensor(img2).permute(2, 0, 1)
        img3 = torch.tensor(img3).permute(2, 0, 1)
        img4 = torch.tensor(img4).permute(2, 0, 1)

        return {
                'image1': img1.float(),
                'image2': img2.float(),
                'offset_image1': img3.float(),
                'augmented_image1': img4.float(),
                'time_sort_label': float(frame_index1 < frame_index2)}

    def num_channels(self):
        return len(self.channels)



import torch
import kwcoco
import os
import ubelt as ub
import kwimage
from .time_sort_module import time_sort


def extract_features(checkpoint,
                     data_folder,
                     kwcoco_file,
                     output_kwcoco,
                     output_folder,
                     image_ids,
                     sensor,
                     panchromatic,
                     device='cuda'):
    """
    Function for extracting features given kwcoco reference to data and
    annotations. Output is copy of input kwcoco file with path towards saved
    pixel-wise features for image_id x saved under
    dset.imgs[x]['time_sort_features'].

    Args:
        checkpoint:  Path to checkpoint of lightning module. Default is UNet base trained on image sorting into before/after.
        data_folder:  Path to dvc repo
        kwcoco_file:  Path to kwcoco file with data annotations
        output_kwcoco:  Destination of output kwcoco file. set to same path as kwcoco_file to simply add paths to feature tensors to the existing file
        output_folder:  destination for feature tensors, stored as .pt files
        image_ids:  Set of image ids (corresponding to image ids in kwcoco_file) from which to extract features. image_ids from non-specified sensors will be skipped. Set to 0 to include all available images.
        sensor:  Choose from S2, LC, or WV. Note: with default checkpoint, only S2 (3 channel) images can be processed
        panchromatic: Set to True to return panchromatic (single channel) WV images where applicable. Otherwise 8 channel images will be returned.

    """
    extractor = time_sort.load_from_checkpoint(
        checkpoint, map_location='cuda').to(device)

    input_dset = kwcoco.CocoDataset.coerce(kwcoco_file)
    output_dset = input_dset.copy()

    if not image_ids:
        # include all available images if none are specified
        image_ids = list(output_dset.index.imgs.keys())

    # Only take images that match the requested sensor
    valid_images = output_dset.images(image_ids)
    flags = [sensor == _ for _ in valid_images.lookup('sensor_coarse')]
    image_ids = valid_images.compress(flags)

    # TODO: could add a subdirectory using some tag associated with the
    # model to differentiate between features from different trained models
    os.makedirs(output_folder, exist_ok=True)

    # TODO: prediction would be faster with a dataset that loaded images
    # in the background while the GPU was predicting.

    # TODO: prediction will likely need to be done on a sliding window

    for gid in ub.ProgIter(image_ids, 'Process image'):

        img = output_dset.index.imgs[gid]
        # The image name should be unique, but if it does not exist, then
        # we have to get creative
        name = img.get('name', None)
        if name is None:
            name = 'timefeat_{:06d}'.format(gid)
        # Construct the filepath we will save the features to
        feature_fpath = os.path.join(output_folder, name + '.tif')

        # TODO: ensure the correct channels and scale wrt to the model are used
        delayed_image = output_dset.delayed_load(gid)
        im = delayed_image.finalize()

        # TODO: Ensure normalization is the same as in training
        # This should be accomplished by storing that info with the model
        image = torch.from_numpy(im.astype('float32')).to(device)

        if len(image.shape) < 3:
            image = image.unsqueeze(-1)

        image = image.permute(2, 0, 1)
        image = image.unsqueeze(0)

        batch_features, _, _, _ = extractor(image, image, 'x', 'x')

        # Assume batch size of 1
        item_features = batch_features[0]

        item_features_np = item_features.data.cpu().numpy().transpose(1, 2, 0)

        height, width, num_bands = item_features_np.shape

        # The input to the network is in "video-space", and the output is given
        # in the same "video-space" space. The output is going to be added as a
        # new auxiliary channel(s) to the image, so we need to specify the warp
        # from auxiliary space to image space, because auxiliary space in this
        # case is video space, we can use the inverse of the image-to-video
        # transform in the image dictionary.
        warp_img_to_vid = kwimage.Affine.coerce(img.get('warp_img_to_vid', None))
        warp_aux_to_img = warp_img_to_vid.inv()

        # TODO: need to come up with a channel code to represent this.
        # currently this could be done by any random 64 codes separated by
        # pipes but we may want to update kwcoco to be nicer in the way
        # it handles larger numbers of channels
        quick_chan_codes = ['UKy{:02d}'.format(i) for i in range(num_bands)]
        channels = '|'.join(quick_chan_codes)

        # Write the data to disk
        kwimage.imwrite(feature_fpath, item_features_np, backend='gdal', space=None)

        # Register the data in the output kwcoco manifest
        _temp_add_auxiliary(output_dset, gid, feature_fpath, width, height,
                            warp_aux_to_img, channels, num_bands)

    output_dset.fpath = output_kwcoco
    print('Write to output_dset.fpath = {!r}'.format(output_dset.fpath))
    output_dset.dump(output_dset.fpath, newlines=True)


def _temp_add_auxiliary(self, gid, fpath, width, height, warp_aux_to_img, channels, num_bands):
    """
    Adds an auxiliary file to an image.

    Temporary function while the kwcoco API is finalized
    """
    aux = {
        'file_name': fpath,
        'width': width,
        'height': height,
        'warp_aux_to_img': kwimage.Affine.coerce(warp_aux_to_img).concise(),
        'channels': channels,
        'num_bands': num_bands,
    }
    # lookup the image you want to add to
    img = self.index.imgs[gid]
    # Ensure there is an auxiliary image list
    auxiliary = img.setdefault('auxiliary', [])
    # Add the auxiliary information to the image
    auxiliary.append(aux)
    self._invalidate_hashid()


def main():
    # TODO: this should be broken out into a function.
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument(
        '--checkpoint',
        default='logs/drop0_sort/train_vidio_1/2021-05-14/default/version_7/checkpoints/epoch=97-step=195.ckpt')  # change to some shared place

    # drop0_aligned dataset arguments
    parser.add_argument(
        '--panchromatic',
        help='set flag for using panchromatic landsat imagery',
        action='store_true')
    parser.add_argument(
        '--sensor',
        type=str,
        help='Choose from WV, LC, or S2',
        default='S2')  # with default checkpoint, we must use RGB images
    parser.add_argument(
        '--data_folder',
        help='path to dvc on local machine',
        default='/localdisk0/SCRATCH/watch/smart_watch_dvc/drop0_aligned/')

    parser.add_argument(
        '--dataset',
        help='kwcoco file with dataset',
        default='/localdisk0/SCRATCH/watch/smart_watch_dvc/drop0_aligned/data.kwcoco.json')

    parser.add_argument(
        '--output_kwcoco',
        help='Filename to save output kwcoco file. Can replace old version.',
        default='/localdisk0/SCRATCH/watch/drop0_features/data_uky_time_sort_features.kwcoco.json')

    parser.add_argument(
        '--output_folder',
        help='Folder to store output feature tenors as .pt files',
        default='/u/eag-d1/scratch/ben/drop0_features/tensors')

    parser.add_argument(
        '--image_ids',
        nargs='+',
        type=int,
        help='Set to 0 for all available images. Otherwise take list of image ids for processing. Images from non-matching sensors will be automatically skipped.',
        default=0)
    parser.add_argument('--device', type=str, default='cuda')

    args = parser.parse_args()

    extract_features(checkpoint=args.checkpoint,
                     data_folder=args.data_folder,
                     kwcoco_file=args.dataset,
                     output_kwcoco=args.output_kwcoco,
                     output_folder=args.output_folder,
                     image_ids=args.image_ids,
                     sensor=args.sensor,
                     panchromatic=args.panchromatic,
                     device=args.device
                     )


if __name__ == '__main__':
    """
    CommandLine:
        WATCH_DATA_DPATH=$(geowatch_dvc)

    """
    main()



import torch
import torch.nn as nn

from watch.tasks.rutgers_material_change_detection.models.base_model import BaseModel
from watch.tasks.rutgers_material_change_detection.models.unet import UNetDecoder, UNetEncoder


class SiameseDifference(BaseModel):
    def __init__(
        self,
        input_size=256,
        num_channels=13,
        threshold=1.3,
        feat_layer=2,
        backbone_name="resnet18",
        pretrain=False,
        mode="bilinear",
        decoder_type="transpose_conv",
    ):

        input_size = [input_size, input_size]
        super(SiameseDifference, self).__init__(input_size)

        self._mode = mode
        self._pretrain = pretrain
        self._threshold = threshold
        self.decoder_type = decoder_type
        self._num_channels = num_channels
        self._feat_layer = "layer" + str(feat_layer)
        self._backbone_name = backbone_name.lower()

        self.build()

    def build(self):

        # Build encoder network to get features.
        if self._backbone_name[:6] == "resnet":
            # Get feature extractor network (resnet).
            self.encoder, feat_dim = self.load_resnet_model(
                self._backbone_name, n_input_channels=self._num_channels, pretrained=self._pretrain
            )
        elif self._backbone_name == "unet":
            assert self._backbone_name == "unet"
            self.encoder = UNetEncoder(n_channels=self._num_channels)
            feat_dim = 512

        # Build decoder network to upsample features.
        if self.decoder_type == "upsample":
            self.decoder = self.build_upsample_conv_decoder(
                height=self._input_size[0],
                width=self._input_size[1],
                feat_channels=feat_dim,  # Features are concatenated
                out_channels=2,
            )
        elif self.decoder_type == "transpose_conv":
            self.decoder = self.build_transpose_conv_decoder(
                height=self._input_size[0],
                width=self._input_size[1],
                feat_channels=feat_dim,  # Features are concatenated
                out_channels=2,
            )
        elif self.decoder_type == "unet":
            assert self._backbone_name == "unet"
            self.decoder = UNetDecoder(n_classes=2, channel_factor=2)

    def forward(self, data):
        """Computes a forward pass of model.

        Compare the difference in feature activation between two images and then resize to the input image shape.

        Args:
            data (dict): A dictionary that must contain an 'images' key. The 'images' key corresponds to a tensor of
                shape [batch_size, num_images, num_channels, height, width].

        Returns:
            output (dict): A dictionary containing model predictions. Return dictionary with 'change_pred' key containing
                a int32 tensor of shape [batch_size, 1, height, width].
        """
        images = data["images"]

        # Expecting a number of images dimension
        assert len(images.shape) == 5

        B, N, C, H, W = images.shape

        # The number of images should be equal or more than two
        assert N >= 2

        # Get first and last image in input data
        ref_image = images[:, 0, ...]
        query_image = images[:, -1, ...]

        # Pass images through pretrained network
        ref_feats = self.model.forward_feat(ref_image)[self._feat_layer]
        query_feats = self.model.forward_feat(query_image)[self._feat_layer]

        # Compute absolute difference between backbone features
        diff_feats = abs(ref_feats - query_feats)  # [B, c, h, w];  h != H & w != W, c != C
        diff_feats, _ = torch.max(diff_feats, dim=1)  # [B, h, w]
        diff_feats = diff_feats.unsqueeze(1)  # [B, 1, h, w]

        # Resize features to original input size -> [B, 1, H, W]
        diff_feats = nn.functional.interpolate(diff_feats, size=(H, W), mode=self._mode, align_corners=True)

        # Compute change prediction
        # Compare difference activation to threshold
        change_pred = diff_feats > self._threshold

        # Convert from bool to integer (int32)
        change_pred = change_pred.int()

        return change_pred

    # def encode(self, ref_frame, query_frame):
    #     """[summary]

    #     Args:
    #         ref_frame ([type]): [description]
    #         query_frame ([type]): [description]

    #     Returns:
    #         (): TODO
    #     """
    #     # Pass images through pretrained network
    #     ref_feats = self.model.forward_feat(ref_frame)[self._feat_layer]
    #     query_feats = self.model.forward_feat(query_frame)[self._feat_layer]

    #     # Compute absolute difference between backbone features
    #     diff_feats = ref_feats - query_feats  # [c, h, w];  h != H & w != W, c != C
    #     change_feats = self.final_change_conv(diff_feats)

    #     return change_feats

    def encode(self, ref_frame, query_frame):
        # Pass images through pretrained network
        # TODO: Clean logic up.
        if self._backbone_name[:6] == "resnet":
            ref_feats = self.encoder.forward_feat(ref_frame)[self._feat_layer]
            query_feats = self.encoder.forward_feat(query_frame)[self._feat_layer]
        else:
            ref_feats = self.encoder(ref_frame)
            query_feats = self.encoder(query_frame)

        if self._backbone_name[:6] == "resnet":
            # Concatenate features along channel dimension.
            # Features shape: [batch_size, n_channels, height, width]
            feats = query_feats - ref_feats
        else:
            feats = []
            for i in range(len(ref_feats)):
                feats.append(query_feats[i] - ref_feats[i])

        return feats

    # def decode(self, feats, height, width):
    #     """[summary]

    #     Args:
    #         feats ([type]): [description]

    #     Returns:
    #         (): TODO
    #     """
    #     # Resize features to original input size -> [1, H, W]
    #     output = nn.functional.interpolate(feats, size=(height, width), mode=self._mode, align_corners=True)
    #     output = output.squeeze()

    #     return output

    def decode(self, feats, height, width):
        output = self.decoder(feats)
        output = output.squeeze()
        return output

    def decode_out_features(self, feats, height, width):
        output_feats = self.decoder.get_output_feats(feats)
        return output_feats


if __name__ == "__main__":
    # Create fake input data (cpu)
    data = {"images": torch.zeros(5, 2, 3, 256, 256)}

    # Build model object
    model = SiameseDifference()

    # Pass data through model
    pred = model(data)
    assert "change_pred" in pred.keys()
    assert len(pred["change_pred"].shape) == 4



# flake8: noqa

from PIL import Image
import torchvision.transforms.functional as FT
import torch
import numpy as np
from scipy.spatial import distance
import random
# import torchvision.transforms.functional.pad
# torch.backends.cudnn.enabled = True
# torch.backends.cudnn.deterministic = True
import itertools
from watch.tasks.rutgers_material_seg.utils import utils

# if 1:
#     torch.set_printoptions(precision=6, sci_mode=False)

IMG_EXTENSIONS = ['*.png', '*.jpeg', '*.jpg', '*.npy']


# FIXME: Hard coding mean/std is not a good idea
mean_std = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])


class RandomCrop(object):

    def __init__(self, size, padding=None, pad_if_needed=True, fill=0, padding_mode='constant'):

        self.size = size
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.fill = fill
        self.padding_mode = padding_mode

    @staticmethod
    def get_params(img, output_size):

        w, h = img.size
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, image, mask):

        # pad the width if needed
        if self.pad_if_needed and image.size[0] < self.size[1]:
            image = FT.pad(image, (self.size[1] - image.size[0], 0), self.fill, self.padding_mode)
            mask = FT.pad(mask, (self.size[1] - mask.size[0], 0), self.fill, self.padding_mode)
        # pad the height if needed
        if self.pad_if_needed and image.size[1] < self.size[0]:
            image = FT.pad(image, (0, self.size[0] - image.size[1]), self.fill, self.padding_mode)
            mask = FT.pad(mask, (0, self.size[0] - mask.size[1]), self.fill, self.padding_mode)

        i, j, h, w = self.get_params(image, self.size)
        crop_image = FT.crop(image, i, j, h, w)
        crop_mask = FT.crop(mask, i, j, h, w)

        return crop_image, crop_mask


class SYSUCDDataset(object):
    def __init__(self, root, transforms, split=False):
        self.root = root
        self.transforms = transforms
        self.split = split

        self.images1_root = f"{self.root}/{split}/time1/"
        self.images2_root = f"{self.root}/{split}/time2/"
        self.masks_root = f"{self.root}/{split}/label/"
        self.masks_paths = utils.dictionary_contents(path=self.masks_root, types=['*.png'])

    def __getitem__(self, idx):
        negative_idx = random.choice([i for i in range(0, self.__len__()) if i != idx])

        mask_path = self.masks_paths[idx]
        negative_image_name = self.masks_paths[negative_idx].split('/')[-1].split('.')[0]
        image_name = mask_path.split('/')[-1].split('.')[0]
        img1_path = f"{self.images1_root}/{image_name}.png"
        img2_path = f"{self.images2_root}/{image_name}.png"
        negative_image_path = f"{self.images1_root}/{negative_image_name}.png"

        img1 = Image.open(img1_path).convert("RGB")
        img2 = Image.open(img2_path).convert("RGB")
        negative_img = Image.open(negative_image_path).convert("RGB")
        mask = Image.open(mask_path)  # .convert("L"))

        new_image1 = self.transforms(img1)
        new_image2 = self.transforms(img2)
        new_negative_image = self.transforms(negative_img)

        print(new_image1.dtype)
        # import matplotlib.pyplot as plt
        # plt.imshow(mask)
        # plt.show()
        new_mask = FT.to_tensor(mask)

        outputs = {}
        outputs['visuals'] = {'image1': new_image1, 'image2': new_image2, 'mask': new_mask, 'image_name': image_name}
        outputs['inputs'] = {'image1': new_image1, 'image2': new_image2, 'negative_image': new_negative_image , 'mask': new_mask}

        return outputs

    def __len__(self):
        return len(self.masks_paths)



import argparse
import logging
import json
import os
import sys
import copy
from contextlib import contextmanager
import shutil

from shapely.geometry import shape, mapping


_logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description="Consolidate regions and sites across "
                    "pseudo-live iterations")

    parser.add_argument('region_id',
                        type=str,
                        help="Region ID")
    parser.add_argument("-o", "--outdir",
                        type=str,
                        required=True,
                        help="Output dir for consolidated regions / sites")
    parser.add_argument("-s", "--performer-suffix",
                        type=str,
                        default=None,
                        help="Performer suffix if present, e.g. KIT")
    parser.add_argument("-i", "--iou-threshold",
                        type=float,
                        default=0.5,
                        help="IOU Threshold for determining duplicates"
                             "(default: 0.5)")
    parser.add_argument('ta2_collated_dir_previous',
                        type=str,
                        help="Path to TA2 collated output dir for "
                             "previous iteration")
    parser.add_argument('ta2_collated_dir_current',
                        type=str,
                        help="Path to TA2 collated output dir for "
                             "current iteration")
    parser.add_argument('--just-deconflict',
                        action='store_true',
                        default=False,
                        help="Don't copy previous sites, just deconflict "
                             "current site IDs with respect to previous")

    pseudolive_consolidate(**vars(parser.parse_args()))


def reindex_ids(site_or_region):
    # Reindex "id" values (not "site_id") to ensure there are no
    # duplicates; modifies in place.  Only needed for sites or region
    # files that we merge / modify
    for new_id, feat in enumerate(site_or_region.get('features', ())):
        feat['id'] = str(new_id)

    return site_or_region


@contextmanager
def _yield_first_feature(site_or_region, type_):
    # Assumes there's only a single 'site' feature
    for feat in site_or_region.get('features', ()):
        if feat['properties']['type'] == type_:
            yield feat
            return


def _load_region_data(region_id, ta2_collated_dir, performer_suffix=None):
    if performer_suffix is not None:
        region_path = os.path.join(
            ta2_collated_dir, 'region_models',
            "{}_{}.geojson".format(region_id, performer_suffix))
    else:
        region_path = os.path.join(
            ta2_collated_dir, 'region_models',
            "{}.geojson".format(region_id))

    with open(region_path) as f:
        return json.load(f)


def _load_site_data(site_id, ta2_collated_dir, performer_suffix=None):
    if performer_suffix is not None:
        site_path = os.path.join(
            ta2_collated_dir, 'site_models',
            "{}_{}.geojson".format(site_id, performer_suffix))
    else:
        site_path = os.path.join(
            ta2_collated_dir, 'site_models',
            "{}.geojson".format(site_id))

    with open(site_path) as f:
        return json.load(f)


def pseudolive_consolidate(region_id,
                           ta2_collated_dir_previous,
                           ta2_collated_dir_current,
                           outdir,
                           iou_threshold,
                           performer_suffix=None,
                           just_deconflict=False):
    os.makedirs(os.path.join(outdir, 'region_models'), exist_ok=True)
    os.makedirs(os.path.join(outdir, 'site_models'), exist_ok=True)

    previous_region = _load_region_data(
        region_id, ta2_collated_dir_previous, performer_suffix)
    current_region = _load_region_data(
        region_id, ta2_collated_dir_current, performer_suffix)

    previous_sites_by_id = {f['properties']['site_id']: f
                            for f in previous_region.get('features', ())
                            if f['properties']['type'] == 'site_summary'}
    current_sites_by_id = {f['properties']['site_id']: f
                           for f in current_region.get('features', ())
                           if f['properties']['type'] == 'site_summary'}

    overlaps = []
    for pss in previous_sites_by_id.values():
        psid = pss['properties']['site_id']
        previous_sites_by_id[psid] = pss

        for css in current_sites_by_id.values():
            csid = css['properties']['site_id']
            current_sites_by_id[csid] = css

            iou = compute_iou(pss, css)
            if iou >= iou_threshold:
                overlaps.append((iou, psid, csid))

    sorted_overlaps = sorted(overlaps, key=lambda o: o[0], reverse=True)

    output_region = copy.deepcopy(previous_region)
    # Strip out site summaries from previous region to create initial
    # new region
    output_region['features'] = [f for f in output_region.get('features', ())
                                 if f['properties']['type'] != 'site_summary']
    with _yield_first_feature(output_region, type_='region') as outr:
        with _yield_first_feature(current_region, type_='region') as cr:
            outr['properties']['end_date'] = cr['properties']['end_date']

    # TODO: Move this function to top-level scope
    def _combine_sites(psid, csid):
        pss = previous_sites_by_id[psid]
        css = current_sites_by_id[csid]

        new_site_summary = copy.deepcopy(pss)
        new_site_summary['properties']['end_date'] =\
            css['properties']['end_date']

        new_site_geometry = mapping(
            shape(pss['geometry']).union(shape(css['geometry'])))
        new_site_summary['geometry'] = new_site_geometry

        # Merge sites
        prev_site = _load_site_data(
            psid, ta2_collated_dir_previous, performer_suffix)
        curr_site = _load_site_data(
            csid, ta2_collated_dir_current, performer_suffix)

        new_site = copy.deepcopy(prev_site)
        # Set the end date of the previous (to be re-used site to the
        # end of the current matching site)
        with _yield_first_feature(new_site, type_='site') as nssf:
            with _yield_first_feature(curr_site, type_='site') as cssf:
                nssf['properties']['end_date'] =\
                    cssf['properties']['end_date']

            nssf['geometry'] = new_site_geometry

        # TODO: Update "phase_transition_days" of last site "observation"?
        new_site['features'].extend(
            [s for s in curr_site.get('features', ())
             if s['properties']['type'] == 'observation'])

        reindex_ids(new_site)

        return new_site_summary, new_site

    # TODO: Move this function to top-level scope
    def _deconflict_new_site(psid, csid):
        # Just copy new site over and re-use previous ID
        css = current_sites_by_id[csid]

        # Use current instead of previous as starting point
        new_site_summary = copy.deepcopy(css)
        new_site_summary['properties']['site_id'] = psid

        curr_site = _load_site_data(
            csid, ta2_collated_dir_current, performer_suffix)

        # Again starting with current instead of previous site
        new_site = copy.deepcopy(curr_site)
        # Set the end date of the previous (to be re-used site to the
        # end of the current matching site)
        with _yield_first_feature(new_site, type_='site') as nssf:
            nssf['properties']['site_id'] = psid

        return new_site_summary, new_site

    matched_previous_sites = set()
    matched_current_sites = set()
    for iou, psid, csid in sorted_overlaps:
        # Only doing one-to-one site matching for now
        if(psid in matched_previous_sites  # noqa: E275
           or csid in matched_current_sites):
            continue

        print("Matched site {} (previous) to {} (current)"
              " with {:.2f} IOU".format(psid, csid, iou))

        matched_previous_sites.add(psid)
        matched_current_sites.add(csid)

        if just_deconflict:
            new_site_summary, new_site = _deconflict_new_site(psid, csid)
        else:
            new_site_summary, new_site = _combine_sites(psid, csid)

        output_region['features'].append(new_site_summary)

        if performer_suffix is None:
            output_path = os.path.join(
                outdir, 'site_models', "{}.geojson".format(psid))
        else:
            output_path = os.path.join(
                outdir, 'site_models',
                "{}_{}.geojson".format(psid, performer_suffix))

        with open(output_path, 'w') as f:
            json.dump(new_site, f, indent=2)

    # Keep track of highest id used for previous sites for when we
    # re-issue IDs for current sites later in this function; starting
    # at -1 as the first site we want to issue should have an ID value
    # of 0
    highest_psid_value = -1
    for psid, ps in previous_sites_by_id.items():
        _, _, psid_value = psid.split('_')
        psid_value = int(psid_value)
        if psid_value > highest_psid_value:
            highest_psid_value = psid_value

        if(psid in matched_previous_sites  # noqa: E275
           or just_deconflict):
            # Already added during merging; or just deconflicting current sites
            continue
        else:
            output_region['features'].append(ps)

            if performer_suffix is not None:
                site_src_path = os.path.join(
                    ta2_collated_dir_previous, 'site_models',
                    "{}_{}.geojson".format(psid, performer_suffix))
                site_dst_path = os.path.join(
                    outdir, 'site_models',
                    "{}_{}.geojson".format(psid, performer_suffix))
            else:
                site_src_path = os.path.join(
                    ta2_collated_dir_previous, 'site_models',
                    "{}.geojson".format(psid))
                site_dst_path = os.path.join(
                    outdir, 'site_models',
                    "{}.geojson".format(psid))

            shutil.copy(site_src_path, site_dst_path)

            print("Copied previous site: {}".format(psid))

    # Re-issue site_ids for "current" sites (that were not merged)
    next_site_id_value = highest_psid_value + 1
    for csid, cs in current_sites_by_id.items():
        if csid in matched_current_sites:
            continue
        else:
            updated_site_id = "{}_{:0>4}".format(region_id, next_site_id_value)
            next_site_id_value += 1
            cs['properties']['site_id'] = updated_site_id

            output_region['features'].append(cs)

            curr_site_data = _load_site_data(
                csid, ta2_collated_dir_current, performer_suffix)

            with _yield_first_feature(curr_site_data, type_='site') as cssf:
                cssf['properties']['site_id'] = updated_site_id

            if performer_suffix is not None:
                site_dst_path = os.path.join(
                    outdir, 'site_models',
                    "{}_{}.geojson".format(updated_site_id, performer_suffix))
            else:
                site_dst_path = os.path.join(
                    outdir, 'site_models',
                    "{}.geojson".format(updated_site_id))

            with open(site_dst_path, 'w') as f:
                json.dump(curr_site_data, f, indent=2)

            print("Copied current site: {} and renamed to: {}".format(
                csid, updated_site_id))

    if performer_suffix is not None:
        region_dst_path = os.path.join(
            outdir, 'region_models',
            "{}_{}.geojson".format(region_id, performer_suffix))
    else:
        region_dst_path = os.path.join(
            outdir, 'region_models',
            "{}.geojson".format(region_id))

    reindex_ids(output_region)
    with open(region_dst_path, 'w') as f:
        json.dump(output_region, f, indent=2)

    print("Wrote consolidated output files to: {}".format(outdir))


def compute_iou(site_summary_1, site_summary_2):
    """
    Addapted from MITRE Corp's Smart Nifi implementation
    https://smartgitlab.com/infrastructure/smart-nifi/-/blob/main/python_scripts/smart_nifi/consolidate_split.py

    Compare two sites and determine if they represent "roughly" the same area.

    this IS going to fail under the following circumstances:
        1) polar regions
        2) sites that cross the 180DEG line

    :param shape1: A Site that is trusted
    :type shape1: a site_summary geojson
    :param shape2: A Site that may be a duplicate
    :type shape2: a site_summary geojson
    :return: iou
    :rtype: float
    """

    # first let's do BBOX checks, because those are incredibly easy, and will
    # rule out 95% of the cases
    shape1 = shape(site_summary_1["geometry"])
    shape2 = shape(site_summary_2["geometry"])

    if shape1.bounds[0] > shape2.bounds[2]:
        _logger.debug(
            "Site %s is strictly east of %s",
            site_summary_2["properties"]["site_id"],
            site_summary_1["properties"]["site_id"],
        )
        return 0.0
    if shape1.bounds[2] < shape2.bounds[0]:
        _logger.debug(
            "Site %s is strictly west of %s",
            site_summary_2["properties"]["site_id"],
            site_summary_1["properties"]["site_id"],
        )
        return 0.0
    if shape1.bounds[1] > shape2.bounds[3]:
        _logger.debug(
            "Site %s is strictly south of %s",
            site_summary_2["properties"]["site_id"],
            site_summary_1["properties"]["site_id"],
        )
        return 0.0
    if shape1.bounds[3] < shape2.bounds[1]:
        _logger.debug(
            "Site %s is strictly north of %s",
            site_summary_2["properties"]["site_id"],
            site_summary_1["properties"]["site_id"],
        )
        return 0.0

    # now that we've filtered most of the cases out, let's do the hard work
    # of checking polygon intersections.

    intersection_area = shape1.intersection(shape2).area
    # probably rare, but worth a check!
    if intersection_area == 0:
        _logger.debug(
            "Site %s does not intersect %s",
            site_summary_2["properties"]["site_id"],
            site_summary_1["properties"]["site_id"],
        )
        return 0.0

    union_area = shape1.union(shape2).area
    if union_area == 0.0:
        return 0.0

    iou = intersection_area / union_area
    _logger.debug(
        "IOU of %s:%s is %0.4f",
        site_summary_2["properties"]["site_id"],
        site_summary_1["properties"]["site_id"],
        iou
    )
    return iou


if __name__ == "__main__":
    sys.exit(main())
